#<: 'config.pl' v0.003 @ 2012-10-07 00:18:56 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

use strict;
use Config::IniHash;

my $debug	= 0;
my $logtxt	= sub {};
my $cfgdir	= '/etc/wtdns';
my $config	= "$cfgdir/wtdns.ini";
my %defaults = (
	''			=> {
		domain		=> 'ddns.example.net',
		datadir		=> '/var/lib/wtdns',
		hasher		=> '/usr/bin/wtdnshash',
		fallback	=> '99m',
		purge		=> '30d',
	},
	fallbacks	=> {
		A			=> '0.1.2.3',
		AAAA		=> '::0.1.2.3',
	},
);

my $yes	= qr/^(?:yes|true|0*[1-9]\s*|on)$/i;
my $no	= qr/^(?:no|false|0+|off)$/i;

my $dbgp	= '';
my $hasher	= '';
my $hashem	= 0;
my $cfg = {};
my $cfgstamp = 0;
my $old = undef;
my $ancient = undef;

sub access_control_lists {
	my %access = ();
	foreach my $sk (keys %{$cfg}) {
		next unless ($sk =~ /^\s*(?:acl|access|user|system)\s*(?:[:=,;]+\s*(.*?))\s*$/);
		my $ids = $1;
		$ids =~ s/^[\s:=,;]+//;
		$ids =~ s/[\s:=,;]+$//;
		my @ids = ($ids eq '') ? () : split(/[\s:=,;]+/,$ids);
		push @ids, $cfg->{$sk}->{id} if ($cfg->{$sk}->{id});
		my @dhs = ();
		foreach my $hn (('host','hosts')) {
			next unless ($cfg->{$sk}->{$hn});
			if (ref($cfg->{$sk}->{$hn}) eq 'ARRAY') {
				push @dhs, @{$cfg->{$sk}->{$hn}};
			} else {
				push @dhs, split(/[\s,;]+/,$cfg->{$sk}->{$hn});
			}
		}
		@dhs = map { $_ =~ s/^\s+//; $_ =~ s/\s+$//; $_ ? $_ : () } @dhs;
		my $dhr = undef;
		if (@dhs) {
			my %dhs = map { $_ .= '.'.$cfg->{''}->{domain} unless ($_ =~ /\./); $_ =~ s/\./\\./g; $_ =~ s/\?/[-_a-zA-Z0-9]/g; $_ =~ s/\*/[-_a-zA-Z0-9]+/; ($_,1) } @dhs;
			$dhr = join('|',keys %dhs);
			$dhr = qr/^($dhr)$/;
			my %flg = ();
			if (defined($cfg->{$sk}->{trust}) && $cfg->{$sk}->{trust} =~ $yes) {
				$flg{I} ++;
			}
			if (defined($cfg->{$sk}->{trust}) && $cfg->{$sk}->{ignore} =~ $yes) {
				$flg{i} ++;
			}
			if ($cfg->{$sk}->{parameters}) {
				foreach my $pv (split(/[^a-z0-9]+/,lc($cfg->{$sk}->{parameters}))) {
					if ($pv eq 'distrust') {
						delete $flg{I};
					} elsif ($pv eq 'trust') {
						$flg{I} ++;
					} elsif ($pv eq 'ignore') {
						$flg{i} ++;
					} elsif ($pv eq 'check') {
						delete $flg{i};
					}
				}
			}
			foreach my $id (@ids) {
				$access{$id} = {
					f => \%flg,
					h => $dhr,
				};
			}
		}
	}
	return wantarray ? %access : \%access;
}

sub load_config {
	$cfg = {};
	return -1 unless (-e $config);
	$cfgstamp = join('|',(stat(_))[1],(stat(_))[7],(stat(_))[9],(stat(_))[10]);
	eval { $cfg = ReadINI($config,systemvars=>0,case=>'tolower',allowmultiple=>{'*'=>['host','group']}); };
	foreach my $k (keys %defaults) {
		$cfg->{$k} = {} unless ($cfg->{$k});
		foreach my $o (keys %{$defaults{$k}}) {
			$cfg->{$k}->{$o} = $defaults{$k}->{$o} unless (defined($cfg->{$k}->{$o}));
		}
	}
	$debug	= (!defined($cfg->{''}->{debug})) ? $debug : ($cfg->{''}->{debug} =~ $yes) ? 1 : ($cfg->{''}->{debug} =~ $no) ? 0 : $debug;
	$dbgp	= $debug ? '{DEBUG} ' : '';
	$hashem = (!$cfg->{''}->{hasher})			? 0 :
			  (!defined($cfg->{''}->{hashing}))	? 1 :
			  ($cfg->{''}->{hashing} =~ $yes)	? 1 :
			  ($cfg->{''}->{hashing} =~ $no)	? 0 : 1;
	$hasher = $cfg->{''}->{hasher};
	$hasher .= ' -uh';
	$hasher .= ' -oh' if (defined($cfg->{''}->{oldhash}) && ($cfg->{''}->{oldhash} =~ $yes));
	$old = undef;
	$ancient = undef;
	eval {
		my $dt = parse_duration($cfg->{''}->{fallback});
		$old = $dt if ($dt > 0);
	};
	eval {
		my $dt = parse_duration($cfg->{''}->{purge});
		$ancient = $dt if ($dt > 0);
	};
	$cfgdir = $config;
	$cfgdir =~ s/[\\\/]*[^\\\/]*$//s;
	return 1;
}

sub reload_config {
	return undef unless (-e $config);
	my $cst = join('|',(stat(_))[1],(stat(_))[7],(stat(_))[9],(stat(_))[10]);
	return 0 if ($cst eq $cfgstamp);
	load_config();
	return 1;
}

my %hashes = ();
sub host_dir {
	my ($host) = @_;
	my @hdir = ();
	my $hdir = $cfg->{''}->{datadir};
	push @hdir, $hdir;
	if ($hashem) {
		unless ($hashes{$host}) {
			return wantarray ? (undef,"wtdnshash error for host '$host' ($!)!") : undef unless (open(F,"$hasher '$host'|"));
			my $hd = <F>;
			close(F);
			return wantarray ? (undef,"wtdnshash error for host '$host' ($hd)!") : undef unless ($hd =~ /^(#?[0-9A-F][0-9A-F])[\r\n]*$/s);
			$hashes{$host} = $1;
		}
		$hdir .= "/$hashes{$host}";	
		push @hdir, $hdir;
	}
	$hdir .= "/$host";
	push @hdir, $hdir;
	return wantarray ? ($hdir,@hdir) : $hdir;
}

load_config();
