#<: 'control.pl' v0.002 @ 2012-10-08 00:52:27 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

use strict;
use Net::DNS::Resolver;

my $ctl = {};
my $ctlstamp = 0;
my $ctlerr = '';
my $ctlcli = undef;

sub control_error {
	($ctlerr) = @_;
	return undef;
}

sub control_pid {
	return 0 unless ($ctl->{pid});
	return 1 unless (-d '/proc');
	my $fh = undef;
	return 0 unless (open($fh,'<',"/proc/$ctl->{pid}/status"));
	while (my $l = <$fh>) {
		next unless ($l =~ /^\s*Name\s*:\s*(\S.*?)[\s\r\n]*$/si);
		close($fh);
		return ($1 eq 'wtdnsd') ? 1 : 0;
	}
	close($fh);
	return 0;
}

sub can_control {
	return 0 unless ($ctl->{port} && $ctl->{domain} && ($ctl->{ipv4udp} || $ctl->{ipv4tcp} || $ctl->{ipv6udp} || $ctl->{ipv6tcp}));
	return 0 unless (control_pid());
	return 1;
}

sub load_control {
	$ctl = {};
	$ctlcli = undef;
	return control_error('No control information available!') unless ($cfg->{''} && $cfg->{''}->{datadir} && (-e "$cfg->{''}->{datadir}/:control"));
	$ctlstamp = join('|',(stat(_))[1],(stat(_))[7],(stat(_))[9],(stat(_))[10]);
	my $fh = undef;
	return control_error("Could not read control info: $!") unless (open($fh,'<',"$cfg->{''}->{datadir}/:control"));
	while (my $l = <$fh>) {
		next unless ($l =~ /^\s*(\S+?)\s*[:=]\s*(.*?)[\s\r\n]*$/s);
		$ctl->{lc($1)} = $2;
	}
	close($fh);
	return can_control();
}

sub reload_control {
	return control_error('No control information available!') unless ($cfg->{''} && $cfg->{''}->{datadir} && (-e "$cfg->{''}->{datadir}/:control"));
	my $cst = join('|',(stat(_))[1],(stat(_))[7],(stat(_))[9],(stat(_))[10]);
	return 0 if ($cst eq $ctlstamp);
	return undef unless (load_control());
	return can_control() ? 1 : -1;
}

sub control_sockets {
	return ((($ctl->{ipv4udp} || $ctl->{ipv4tcp}) ? '127.0.0.1' : ()), (($ctl->{ipv6udp} || $ctl->{ipv6tcp}) ? '::1' : ()));
}

sub control_client {
	return $ctlcli if ($ctlcli);
	unless ($ctl->{port} && $ctl->{domain} && ($ctl->{ipv4udp} || $ctl->{ipv4tcp} || $ctl->{ipv6udp} || $ctl->{ipv6tcp})) {
		$ctlerr = "No control socket available!";
		return undef;
	}
	$ctlcli = Net::DNS::Resolver->new(
		debug			=> $debug,
		recurse			=> 0,
		port			=> $ctl->{port},
		nameservers		=> [control_sockets()],
		tcp_timeout		=> 5,
		retrans			=> 1,
		retry			=> 5,
		dnsrch			=> 0,
		persistent_tcp	=> 1,
		persistent_udp	=> 1,
	);
	unless ($ctlcli) {
		$ctlerr = "DNS Resolver error!";
		return undef;
	}
	$ctlcli->dnssec(1);
	$ctlcli->udppacketsize(20480);
	return $ctlcli;
}

sub query_daemon_status {
	my ($qtype,$qname,$dname) = @_;
	my $dns = control_client();
	return undef unless ($dns);
	$qtype = 'ANY' unless ($qtype);
	unless ($dname) {
		$dname = $qname ? $qname : 'info';
		$dname .= $ctl->{domain};
	}
	my $res = $dns->send($dname,$qtype);
	return {r=>$res,a=>[control_sockets],p=>$ctl->{port}} if ($res);
	$ctlerr = "No answer from daemon!";
	return 0;
}

my $dns = undef;
sub query_daemon_rr {
	my ($qtype,$qname) = @_;
	my $dns = control_client();
	return undef unless ($dns);
	return $dns->query($qname,$qtype);
}

load_control();
