#<: 'dnssec.pl' v0.004 @ 2012-10-08 21:22:02 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

use strict;
use Net::DNS::Resolver;
use Net::DNS::RR::RRSIG;
use Net::DNS::SEC::Private;
use Net::DNS::RR::NSEC;
use Net::DNS::RR;
use Date::Format;
use MIME::Base64;

my %dnssec_keys = ();
sub dnssec_find_keys {
	my $zone = shift @_;
	my @keyl = map { $_ ? $_ : () } map { ref($_) ? @{$_} : $_ } @_;
	push @keyl, '||';
	while ($zone) {
		push @keyl, $zone;
		$zone =~ s/^[^.]+//s;
		$zone =~ s/^\.+//s;
	}
	push @keyl, ('||','dnssec');
	my %keyc = ();
	foreach my $kw (@keyl) {
		if ($kw eq '||') {
			last if (%keyc);
			next;
		}
		$kw =~ s/\.(ds|key|private)$//;
		my $kv = ($kw =~ /[:\\\/]/) ? "$kw*" : "$cfgdir/$kw*";
		my $fc = 0;
		foreach my $kf (glob($kv)) {
			next unless ($kf =~ s/\.(ds|key|private)$//);
			$keyc{$kf} = {} unless ($keyc{$kf});
			$keyc{$kf}->{lc($1)} ++;
		}
		next if ($fc);
		next if ($kw =~ /[:\\\/]/);
		$kv = "$cfgdir/K$kw*";
		foreach my $kf (glob($kv)) {
			next unless ($kf =~ s/\.(ds|key|private)$//);
			$keyc{$kf} = {} unless ($keyc{$kf});
			$keyc{$kf}->{lc($1)} ++;
		}
	}
	my $now = time();
	if (%dnssec_keys) {
		my $old = $now - 3600;
		foreach my $key (keys %dnssec_keys) {
			if ($dnssec_keys{$key}->{ut} < $old) {
				delete $dnssec_keys{$key};
			}
		}
	}
	@keyl = ();
	my @save = ();
	foreach my $key (sort { length($a) <=> length($b) } keys %keyc) {
		next unless ($keyc{$key}->{ds} && $keyc{$key}->{key} && $keyc{$key}->{private});
		my $file = "$key.private";
		unless (-f $file) {
			delete $dnssec_keys{$key} if (exists $dnssec_keys{$key});
			next;
		}
		my $mod = (stat(_))[9];
		if ($dnssec_keys{$key} && ($dnssec_keys{$key}->{mt} == $mod)) {
			$dnssec_keys{$key}->{ut} = $now;
			push @keyl, {k=>$dnssec_keys{$key}->{pk},f=>$key};
		} else {
			$logtxt->('Loading',$key);
			my $prv = undef;
			eval { $prv = Net::DNS::SEC::Private->new($file); };
			if ($prv) {
				$dnssec_keys{$key} = {mt=>$mod,pk=>$prv,ut=>$now};
				my $ke = {k=>$prv,f=>$key};
				push @keyl, $ke;
				push @save, $ke;
			} elsif ($dnssec_keys{$key}) {
				$dnssec_keys{$key}->{ut} = $now;
				push @keyl, {k=>$dnssec_keys{$key}->{pk},f=>$key};
			}
		}
	}
	if (@save) {
		dnssec_save_dnskeys($zone,{l=>\@save,z=>$zone});
		dnssec_save_dnskey_digests($zone,{l=>\@save,z=>$zone});
	}
	return {l=>\@keyl,z=>$zone};
}

sub dnssec_sign_rrset {
	my ($zone,$keyr,$rrset) = @_;
	$keyr = dnssec_find_keys($zone,$keyr) unless (ref($keyr) eq 'HASH');
	return wantarray ? () : undef unless ($keyr);
	my @rrs = ();
	foreach my $k (@{$keyr->{l}}) {
		my $rrsig = undef;
		eval { $rrsig = Net::DNS::RR::RRSIG->create($rrset,$k->{k},sigex=>time2str("%Y%m%d%H%M%S",time()+(7305*24*60*60))); };
		push @rrs, $rrsig if ($rrsig);
	}
	return wantarray ? @rrs : \@rrs;
}

sub dnssec_create_nsecs {
	my ($zone) = @_;
	return wantarray ? () : undef unless ($zone);
	$logtxt->('Listing',$zone);
	my $qres = query_daemon_rr('PTR',"*.$zone");
	return wantarray ? () : undef unless ($qres);
	my @names = ();
	my $rc = 0;
	foreach my $qans ($qres->answer) {
		next unless ($qans->type eq 'PTR');
		$rc ++;
		next unless ($qans->ptrdname =~ /^[^.]+\.$zone$/i);
		push @names, $qans->ptrdname;
	}
	return wantarray ? () : 0 unless ($rc);
	@names = sort { lc($a) cmp lc($b) } @names;
	my @nsecs = ();
	my $name = $names[0];
	$name =~ s/^[^.]*/!/s;
	unshift @names, $name;
	for (my $i=0;$i<@names;$i++) {
		my $nsec = undef;
		if ($i < $#names) {
			my $qres = query_daemon_rr('ANY',$names[$i+1]);
			my %types = ('NSEC'=>1);
			foreach my $qans ($qres ? $qres->answer: ()) {
				$types{$qans->type} ++;
			}
			eval { $nsec = Net::DNS::RR->new(type=>'NSEC',name=>$names[$i],nxtdname=>$names[$i+1],typelist=>join(' ',keys %types)); };
		} else {
			$name = $names[$i];
			$name =~ s/^[^.]*/~/s;
			eval { $nsec = Net::DNS::RR->new(type=>'NSEC',name=>$names[$i],nxtdname=>$name,typelist=>'NSEC RRSIG'); };
		}
		push @nsecs, $nsec if ($nsec);
	}
	return wantarray ? @nsecs : \@nsecs;
}

sub dnssec_sign_name {
	my ($zone,$keyr,$name) = @_;
	$zone = $name unless ($zone);
	$logtxt->('Fetching',$name);
	my $qres = query_daemon_rr('ANY',$name);
	return wantarray ? () : 0 unless ($qres);
	my %rrsets = ();
	foreach my $qans ($qres->answer) {
		next if ($qans->type eq 'RRSIG' || $qans->type eq 'NSEC');
		$rrsets{$qans->type} = [] unless ($rrsets{$qans->type});
		push @{$rrsets{$qans->type}}, $qans;
	}
	my @rrsigs = ();
	foreach my $type (keys %rrsets) {
		$logtxt->('Signing',$type,$name);
		push @rrsigs, dnssec_sign_rrset($zone,$keyr,$rrsets{$type});
	}
	return wantarray ? @rrsigs : \@rrsigs;
}

sub dnssec_sign_names {
	my ($zone,$keyr,@names) = @_;
	@names = map { $_ ? $_ : () } map { ref($_) ? @{$_} : $_ } @names;
	my @rrsigs = ();
	foreach my $name (@names) {
		push @rrsigs, dnssec_sign_name($zone,$keyr,$name);
	}
	return wantarray ? @rrsigs : \@rrsigs;
}

sub dnssec_sign_zone {
	my ($zone,$keyr) = @_;
	$keyr = dnssec_find_keys($zone,$keyr) unless (ref($keyr) eq 'HASH');
	return wantarray ? () : undef unless ($keyr);
	$logtxt->('Listing',$zone);
	my $qres = query_daemon_rr('PTR',"*.$zone");
	return wantarray ? () : 0 unless ($qres);
	my @rrsigs = ();
	foreach my $qans ($qres->answer) {
		next unless ($qans->type eq 'PTR');
		push @rrsigs, dnssec_sign_name($zone,$keyr,$qans->ptrdname);
	}
	return wantarray ? @rrsigs : \@rrsigs;
}

sub dnssec_save_rrs {
	my ($zone,$keyr,@rrs) = @_;
	@rrs = map { $_ ? $_ : () } map { (ref($_) && ref($_) !~ /::/) ? @{$_} : $_ } @rrs;
	my %err = ();
	my $c = 0;
	my %typc = ();
	foreach my $rr (@rrs) {
		$typc{$rr->type} ++;
	}
	$logtxt->("Saving",$zone,join(' ',map { "$_:$typc{$_}" } sort { $typc{$b} <=> $typc{$a} } keys %typc)) if (%typc);
	while (@rrs) {
		my %rrs = ();
		while (@rrs) {
			my $rr = shift @rrs;
			my ($type,$data);
			my $name = $rr->name;
			if ($rr->type eq 'RRSIG') {
				$type = join(';',$rr->typecovered,$rr->type);
				$data = join(';',$rr->algorithm,$rr->labels,$rr->orgttl,$rr->sigexpiration,$rr->siginception,$rr->keytag,$rr->signame,$rr->sig);
			} else {
				$type = $rr->type;
				if ($rr->type eq 'NSEC') {
					my %types = map { $_ ? ($_,1) : () } split(/\s+/,$rr->typelist);
					$data = join(';',$rr->nxtdname,join(',',keys %types),unpack('H*',$rr->typebm));
				} elsif ($rr->type eq 'DNSKEY') {
					$data = join(';',$rr->flags,$rr->protocol,$rr->algorithm,$rr->key);
				} elsif ($rr->type eq 'DS') {
					$data = join(';',$rr->keytag,$rr->algorithm,$rr->digtype,$rr->digest);
				} else {
					next;
				}
			}
			$data =~ s/[\r\n]+//gs;
			next unless ($data && $data !~ /^\s*$/);
			$rrs{$name} = {} unless ($rrs{$name});
			$rrs{$name}->{$type} = {r=>[],d=>[]} unless ($rrs{$name}->{$type});
			push @{$rrs{$name}->{$type}->{r}}, $rr;
			push @{$rrs{$name}->{$type}->{d}}, $data;
		}
		foreach my $name (keys %rrs) {
			foreach my $type (keys %{$rrs{$name}}) {
				my ($did,$file) = update_record($name,$type,join("\n",@{$rrs{$name}->{$type}->{d}},''));
				next if ($did eq '=');
				unless ($did eq '+' || $did eq '%') {
					$err{$ctlerr} ++;
					next;
				}
				$c ++;
				$logtxt->('Updated',$type,$name);
				next if ($type =~ /;RRSIG$/);
				$logtxt->('Signing',$type,$name);
				push @rrs, dnssec_sign_rrset($zone,$keyr,$rrs{$name}->{$type}->{r});
			}
		}
	}
	return ($c,join(';',keys %err));
}

sub dnssec_save_dnskeys {
	my ($zone,$keyr) = @_;
	$keyr = dnssec_find_keys($zone,$keyr) unless (ref($keyr) eq 'HASH');
	return undef unless ($keyr);
	my %data = ();
	foreach my $k (@{$keyr->{l}}) {
		my $fh = undef;
		next unless (open($fh,'<',"$k->{f}.key"));
		while (my $l = <$fh>) {
			next if ($l =~ /^\s*[;:#]/);
			$l =~ s/;.*$//s;
			next unless ($l =~ /^\s*(\S+)(\.\s+IN\s+DNSKEY\s+\d+\s+\d+\s+\d+\s+.*?)[\s\r\n]*$/s);
			my $name = $1;
			$l = $2;
			$name = $zone if ($zone);
			$data{$name} = [] unless ($data{$name});
			my $rr;
			eval { $rr = Net::DNS::RR->new("$name$l"); };
			push @{$data{$name}}, $rr if ($rr && $rr->type eq 'DNSKEY');
		}
	}
	my $c = 0;
	foreach my $name (keys %data) {
		$c += dnssec_save_rrs($name,$keyr,$data{$name});
	}
	return $c;
}

sub dnssec_save_dnskey_digests {
	my ($zone,$keyr) = @_;
	$keyr = dnssec_find_keys($zone,$keyr) unless (ref($keyr) eq 'HASH');
	return undef unless ($keyr);
	my %data = ();
	foreach my $k (@{$keyr->{l}}) {
		my $fh = undef;
		next unless (open($fh,'<',"$k->{f}.ds"));
		while (my $l = <$fh>) {
			next if ($l =~ /^\s*[;:#]/);
			$l =~ s/;.*$//s;
			next unless ($l =~ /^\s*(\S+)(\.\s+IN\s+DS\s+\d+\s+\d+\s+\d+\s+.*?)[\s\r\n]*$/s);
			my $name = $1;
			$l = $2;
			$name = $zone if ($zone);
			$data{$name} = [] unless ($data{$name});
			my $rr;
			eval { $rr = Net::DNS::RR->new("$name$l"); };
			push @{$data{$name}}, $rr if ($rr && $rr->type eq 'DS');
		}
	}
	my $c = 0;
	foreach my $name (keys %data) {
		$c += dnssec_save_rrs($name,$keyr,$data{$name});
	}
	return $c;
}
