#!/usr/bin/perl
#<: 'wtdns-fromlog.pl' v0.011 @ 2012-10-06 23:30:55 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

# Apache:

# LogFormat "t=%{%s}t|T=%{%Y-%m-%d %H:%M:%S}t|s=%A|p=%p|c=%>s|i=%a|u=%u|f=%f|r=%U|q=%q|a=%{User-Agent}i" DDNSUpdate

# t = stamp
# T = time
# s = server ip
# p = server port
# i = client ip
# u = client user
# f = result file
# r = request url
# q = query string
# a = user agent

# ==================================================================

our $VERSION = '0.011';

use strict;
use Date::Format;
use Date::Parse;
use Data::Validate::IP qw(is_ipv4 is_ipv6);
use URL::Encode qw(url_decode);
use Time::Duration::Parse;

require 'config.pl';
my %access = ();
my $old = undef;
sub init_config {
	%access = access_control_lists();
	$old = undef;
	foreach my $dk (('purge','fallback')) {
		next unless ($cfg->{''}->{$dk});
		eval {
			my $dt = parse_duration($cfg->{''}->{$dk});
			$old = $dt if ($dt > 0 && (!defined($old) || $dt < $old));
		};
	}
}
init_config();

require 'logging.pl';
my $logorder = '~!-+=@cu>i$?rqaspft#T';
init_log('wtdnsfl',$logorder);
syslog('info','fromlog %s','started');

require 'usergroup.pl';

require 'update.pl';

my %hashes = ();
sub handle {
	my ($l) = @_;
	$l =~ s/[\r\n]+$//;
	my %req = ();
	if ($l =~ s/\|a=(.*)$//) {
		$req{a} = $1;
	}
	%req = (%req,map { ($_ =~ /^([a-zA-Z])=(.*)$/) ? ($1,$2) : () } split(/\|/,$l));
	return 0 unless (%req);
	return loginf(%req) unless ($req{c} && $req{u} && $req{f} && $req{c} == 200 && $req{f} =~ /[^a-z0-9](?:ok|good|received)(?:\.[a-z]+)?$/i);
	if (reload_config()) {
		syslog('info','reconfigured');
		init_config();
	}
	if ($old) {
		my $st = $req{t} ? $req{t} : $req{T} ? str2time($req{T}) : undef;
		if (defined($st)) {
			my $age = abs(time() - $st);
			return loginf(%req,'!'=>"$age>$old") if ($old && $age > $old);
			#return loginf(%req,'!'=>"$age>$ancient") if ($ancient && $age > $old);
		}
	}
	my $acc = $access{$req{u}};
	return loginf(%req,'!'=>'*') unless ($acc && $acc->{h});
	if ($req{q}) {
		$req{'?'} = $req{q};
		$req{'?'} =~ s/^\?+//;
	} elsif ($req{r} =~ m#^.*[a-zA-Z]/(.*)$#) {
		$req{'?'} = $1;
	}
	$req{'$'} = {};
	if ($req{'?'}) {
		my @v = split(/[;&]+/,$req{'?'});
		while (@v) {
			my $v = shift @v;
			if ($v =~ /^(.+?)=(.+)$/) {
				my $c = lc(url_decode($1));
				$v = url_decode($2);
				$v =~ s/^\s+//;
				$v =~ s/\s+$//;
				$c =~ s/[^a-z0-9]+//;
				next if ($c eq '' || $v eq '');
				$req{'$'}->{$c} = [] unless ($req{'$'}->{$c});
				push @{$req{'$'}->{$c}}, $v;
			} else {
				$v = lc(url_decode($v));
				foreach my $h (split(/[^-_.a-z0-9]+/,$v)) {
					next if ($h eq '');
					$h .= '.'.$cfg->{''}->{domain} unless ($h =~ /\./);
					$req{'@'} = [] unless ($req{'@'});
					push @{$req{'@'}}, $h;
				}
			}
		}
	}
	return loginf(%req) unless ($req{'@'});
	foreach my $h (@{$req{'@'}}) {
		next if ($h =~ $acc->{h});
		$req{'!'} = [] unless ($req{'!'});
		push @{$req{'!'}}, $h;
	}
	return loginf(%req) if ($req{'!'});
	if ($req{'$'}->{myip} && !$acc->{f}->{i}) {
		if ($acc->{f}->{I}) {
			$req{'>'} = {map { ($_,'-') } @{$req{'$'}->{myip}}};
		} else {
			my %i = ();
			foreach my $i (@{$req{'$'}->{myip}}) {
				next if (exists $i{$i});
				$i{$i} = '';
				unless ($i eq $req{i}) {
					$req{'~'} = [] unless ($req{'~'});
					push @{$req{'~'}}, $i;
				}
			}
			return loginf(%req) if ($req{'~'});
			$req{'>'} = \%i;
		}
	} elsif ($req{i}) {
		$req{'>'} = {$req{i}=>''};
	}
	return loginf(%req) unless ($req{'>'} && %{$req{'>'}});
	my $uf = 0;
	foreach my $i (keys %{$req{'>'}}) {
		if (is_ipv4($i)) {
			$req{'>'}->{$i} = 'A';
		} elsif (is_ipv6($i)) {
			$req{'>'}->{$i} = 'AAAA';
		} else {
			$req{'>'}->{$i} = '?';
			$uf ++;
		}
	}
	return loginf(%req) if ($uf);
	my %hh = ();
	$req{'-'} = {};
	$req{'='} = {};
	$req{'+'} = {};
	foreach my $h (@{$req{'@'}}) {
		$hh{$h} = {} unless ($hh{$h});
		my %th = (map { ($req{'>'}->{$_},$_) } keys %{$req{'>'}});
		foreach my $t (keys %th) {
			next if (exists $hh{$h}->{$t});
			$hh{$h}->{$t} ++;
			my ($did,$inf) = update_record($h,$t,$th{$t});
			if ($did eq '=') {
				$req{'='}->{$t} = $th{$t};
			} elsif ($did eq '+' || $did eq '%') {
				$req{'+'}->{$t} = $th{$t};
			} else {
				$req{'-'}->{$h} = $did ? $did : 'u';
				logerr($inf ? $inf : "Error: $h");
			}
		}
	}
	$req{'@'} = [keys %hh];
	if ($req{'+'} && %{$req{'+'}}) {
		$req{''} = 'updated';
	} elsif ($req{'='} && %{$req{'='}}) {
		$req{''} = 'touched';
	} else {
		$req{''} = 'checked';
	}
	loginf(%req);
	return 1;
}

while (my $l = <STDIN>) {
	$l =~ s/[\r\n]+$//;
	handle($l);
}
syslog('info','%s','finished');
