#!/usr/bin/perl
#<: 'wtdns-maintain.pl' v0.006 @ 2012-09-26 18:02:43 UTC :>#

our $VERSION = '0.006';

use strict;
use File::Temp qw(tempfile);
use File::Copy;
use Time::Duration::Parse;

require 'config.pl';
my $old = '';
my $ancient = '';
eval {
	my $dt = parse_duration($cfg->{''}->{fallback});
	$old = $dt if ($dt > 0);
};
eval {
	my $dt = parse_duration($cfg->{''}->{purge});
	$ancient = $dt if ($dt > 0);
};

require 'logging.pl';
my $logorder = 'ntfarcdp';
init_log('wtdnsm',$logorder);
$logstdout = 1 if ($debug);

require 'usergroup.pl';

my @ddirs = ();
unless (opendir(D,$cfg->{''}->{datadir})) {
	syslog('err',"Data dir '%s' error: %s",$cfg->{''}->{datadir},$!);
	print STDERR "Data dir $cfg->{''}->{datadir} error: $!\n";
	exit(2);
}
if ($hashem) {
	while (my $d = readdir(D)) {
		push @ddirs, "$cfg->{''}->{datadir}/$d" if ($d =~ /^#?[0-9A-F][0-9A-F]$/);
	}
} else {
	push @ddirs, $cfg->{''}->{datadir};
}
closedir(D);

my $dex = $cfg->{''}->{domain};
$dex =~ s/\./\\./g;
$dex = qr/\.$dex$/;
my @hdirs = ();
while (my $dd = shift @ddirs) {
	next unless (opendir(D,$dd));
	while (my $d = readdir(D)) {
		next if ($d eq $cfg->{''}->{domain});
		next if ($d =~ /^\./);
		next if ($d =~ /^\*\./);
		next unless ($d =~ $dex);
		push @hdirs, {d=>$dd,n=>$d};
	}
	closedir(D);
}

my $now = time();
my @fall = ();
my @purge = ();
while (my $hdn = shift @hdirs) {
	next unless (-w "$hdn->{d}/$hdn->{n}");
	next unless (opendir(D,"$hdn->{d}/$hdn->{n}"));
	while (my $f = readdir(D)) {
		next if ($f =~ /^\./);
		next unless (-f "$hdn->{d}/$hdn->{n}/$f");
		next unless (-w "$hdn->{d}/$hdn->{n}/$f");
		my $mt = (stat(_))[9];
		my %rfi = %{$hdn};
		$rfi{t} = $f;
		$rfi{a} = $now - $mt;
		$rfi{P} = "$ancient ($cfg->{''}->{purge})" if ($ancient);
		$rfi{F} = "$old ($cfg->{''}->{fallback})" if ($old);
		if ($ancient && ($rfi{a} > $ancient)) {
			$rfi{r} = 'p';
			push @purge, \%rfi;
		} elsif ($old && ($rfi{a} > $old) && $cfg->{fallbacks}->{$f}) {
			if (open(F,'<',"$hdn->{d}/$hdn->{n}/$f")) {
				my $fc = <F>;
				close(F);
				$fc =~ s/[\r\n]+$//;
				if ($fc eq $cfg->{fallbacks}->{$f}) {
					$rfi{c} = $fc;
					$rfi{r} = 'i';
				} else {
					$rfi{f} = $cfg->{fallbacks}->{$f};
					$rfi{r} = 'f';
					$rfi{m} = $mt;
					push @fall, \%rfi;
				}
			} else {
				$rfi{r} = 'd';
			}
		} else {
			$rfi{r} = 'i';
		}
		loginf(%rfi,''=>'found') if ($debug);
	}
	closedir(D);
}

while (my $fi = shift @fall) {
	unless ($debug) {
		my $tmp = join(':','',$fi->{t},$$,'XXXXX');
		my ($th,$tf) = (undef,undef);
		eval { ($th,$tf) = tempfile("$fi->{d}/$fi->{n}/$tmp",UNLINK=>0); };
		unless ($th && $tf) {
			logerr('tempfile error',"$fi->{d}/$fi->{n}/$tmp");
			next;
		}
		eval { chmod(0644,$tf); };
		print $th "$fi->{f}\n";
		close($th);
		eval { utime(time(),$fi->{m},$tf); };
		unless (move($tf,"$fi->{d}/$fi->{n}/$fi->{t}")) {
			logerr('move error',$tf,"$fi->{d}/$fi->{n}/$fi->{t}");
			unlink($tf);
			next;
		}
		eval { chmod(0644,"$fi->{d}/$fi->{n}/$fi->{t}"); };
		eval { utime(time(),$fi->{m},"$fi->{d}/$fi->{n}/$fi->{t}"); };
	}
	loginf(%{$fi},''=>'felled');
}

while (my $pi = shift @purge) {
	unless ($debug) {
		unless (unlink("$pi->{d}/$pi->{n}/$pi->{t}")) {
			logerr('unlink error',"$pi->{d}/$pi->{n}/$pi->{t}");
			next;
		}
		logerr('purge error',"$pi->{d}/$pi->{n}/$pi->{t}") if (-e "$pi->{d}/$pi->{n}/$pi->{t}");
	}
	loginf(%{$pi},''=>'purged');
}
