#!/usr/bin/perl
#<: 'wtdnsadd.pl' v0.010 @ 2012-10-06 23:31:19 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

our $VERSION = '0.010';

use strict;

require 'config.pl';

unless (@ARGV) {
	my $me = $0;
	$me =~ s/^.*[\\\/]//;
	print "$me <host> <RR> <data> [<RR> <data> [...]]\n";
	exit(1);
}

my $host = shift @ARGV;
unless ($host) {
	print STDERR "Need host!\n";
	exit(2);
}
$host .= '.'.$cfg->{''}->{domain} unless ($host =~ /\./);

my %rrs = ();
while (@ARGV) {
	my $type = uc(shift @ARGV);
	$type =~ s/^\s+//;
	$type =~ s/\s+$//;
	unless ($type =~ /^[A-Z]+$/) {
		print STDERR "Bad RR type '$type'!\n";
		exit(2);
	}
	my $data = @ARGV ? shift @ARGV : '';
	$data =~ s/^\s+//;
	$data =~ s/\s+$//;
	if ($data eq '') {
		print STDERR "RR '$type' needs data!\n";
		exit(2);
	}
	$rrs{$type} = [] unless ($rrs{$type});
	push @{$rrs{$type}}, $data;
}
unless (%rrs) {
	print STDERR "Need RRs!\n";
	exit(2);
}

require 'update.pl';

foreach my $type (keys %rrs) {
	my ($did,$inf) = update_record($host,$type,join("\n",@{$rrs{$type}},''));
	if ($did eq '=') {
		print "$dbgp$inf data unchanged\n";
	} elsif ($did eq '%') {
		print "$dbgp$inf updated\n";
	} elsif ($did eq '+') {
		print "$dbgp$inf added\n";
	} else {
		print STDERR "($did) $inf\n";
	}
}
