#!/usr/bin/perl
#<: 'wtdnsinfo.pl' v0.007 @ 2012-10-06 22:01:19 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

our $VERSION = '0.007';

use strict;
use Net::DNS::Resolver;
use Linux::Smaps;

require 'config.pl';
require 'control.pl';
my $cerr = 0;

unless (can_control()) {
	print STDERR "$ctlerr\n";
	$cerr ++;
	exit(1);
}

my %info = ();
my $poso = 1;
my $cerr = 0;

sub get_meminfo {
	unless ($ctl->{pid}) {
		print STDERR "No PID available!\n";
		$cerr ++;
		return undef;
	}
	my %ofni = ();
	eval {
		my $smaps = Linux::Smaps->new($ctl->{pid});
		my @smaps = ();
		eval { push @smaps, $smaps->pss; };
		eval { push @smaps, $smaps->rss; };
		$ofni{memory} = join(' / ',map { $_ ? "$_ KiB" : () } @smaps);
	};
	unless (control_pid()) {
		print STDERR "PID not wtDNSd!\n";
		$cerr ++;
		return 0;
	}
	%info = (%info,%ofni);
	return 1;
}

sub get_daemoninfo {
	my $res = query_daemon_status('ANY','info');
	return $res unless ($res);
	foreach my $inf ($res->{r}->answer) {
		if ($inf->type eq 'PTR') {
			$info{hostname} = $inf->ptrdname;
		} elsif ($inf->type eq 'A' || $inf->type eq 'AAAA') {
			$info{listen} = {} unless ($info{listen});
			$info{listen}->{_} = [] unless (exists $info{listen}->{_});
			push @{$info{listen}->{_}}, $inf->address if ($info{listen}->{_});
		} elsif ($inf->type eq 'SRV') {
			$info{listen}->{$inf->port} = [] unless (exists $info{listen}->{$inf->port});
			if ($info{listen}->{$inf->port}) {
				push @{$info{listen}->{$inf->port}}, $inf->target;
				$info{listen}->{_} = undef;
			}
		} elsif ($inf->type eq 'HINFO') {
			$info{system} = join(' ',$inf->os ? $inf->os : (), $inf->cpu ? '('.$inf->cpu.')' : ());
		} elsif ($inf->type eq 'TXT') {
			if ($inf->txtdata =~ /^(?:[a-z][a-zA-Z0-9]+:\d+)(?:, [a-z][a-zA-Z0-9]+:\d+)*$/) {
				$info{counters} = {} unless ($info{counters});
				foreach my $cs (split(/\s*,\s*/,$inf->txtdata)) {
					next unless ($cs =~ /^([a-z]+)([A-Z][a-zA-Z0-9]*):(\d+)$/);
					$info{counters}->{$1} = {} unless ($info{counters}->{$1});
					$info{counters}->{$1}->{$2} = $3;
				}
			} elsif ($inf->txtdata =~ /^([-: 0-9]+?(?: [-+A-Z0-9]+)?) - ([-: 0-9]+?(?: [-+A-Z0-9]+)?) = ([ 0-9a-z]+?)\s*$/) {
				$info{started} = $1;
				$info{now} = $2;
				$info{uptime} = $3;
			} elsif ($inf->txtdata =~ /^(.*? v?[.0-9]+) \(([-: 0-9]+(?: [-+A-Z0-9]+)?)\)$/) {
				$info{daemon} = $1;
				$info{compiled} = $2;
			} elsif ($inf->txtdata !~ /^[\s\r\n]*$/) {
				$info{text} = [] unless ($info{text});
				push @{$info{text}}, $inf->txtdata;
			}
		}
	}
}

sub print_info {
	my $ps = 0;
	my %fo = map { ($_,$poso++) } reverse qw(daemon compiled system hostname started uptime memory);
	foreach my $fld (sort { $fo{$b} <=> $fo{$a} } keys %info) {
		next if (ref($info{$fld}));
		next if ($fld eq 'now');
		next if ($info{$fld} eq '');
		print "\n" unless ($ps);
		print sprintf("%-10s %s\n",ucfirst("$fld:"),$info{$fld});
		$ps ++;
	}
	if ($info{listen}) {
		my $prf = "Listen:";
		foreach my $prt (sort { $a <=> $b } keys %{$info{listen}}) {
			next unless ($info{listen}->{$prt});
			foreach my $adr (sort @{$info{listen}->{$prt}}) {
				next unless ($adr);
				print "\n" unless ($ps);
				print sprintf("%-10s %s:%u\n",$prf,$adr,$prt);
				$prf = '';
				$ps ++;
			}
		}
	}
	return $ps;
}

sub print_counters {
	my $pa = 0;
	my %sn = (
		'c'  => 'Conn In',
		'm'  => 'Msg  In',
		'r'	 => 'Msg  Out',
		'zi' => 'Size In',
		'zo' => 'Size Out',
		'zm' => 'Size Max',
		'tm' => 'Type In',
		'tr' => 'Type Out',
	);
	my %so = map { ($_,$poso++) } reverse qw(c m r zi zo zm tm tr);
	my %co = map { ($_,$poso++) } reverse qw(Tot Msg UDP TCP IPv4 IPv6 OPT DO Non Big Bad Err);
	foreach my $shw (([qr/^c/,undef],[undef,qr/^[czt]/],[qr/^t/,undef],[qr/^z/,undef])) {
		my $ps = 0;
		if ($info{counters}) {
			print "\n";
			my @css = map { (($shw->[0] && $_ !~ $shw->[0]) || ($shw->[1] && $_ =~ $shw->[1])) ? () : $_ } keys %{$info{counters}};
			my %cfs = map { map { ($_,1) } keys %{$info{counters}->{$_}} } @css;
			my @cfs = sort { ($co{$b} != $co{$a}) ? $co{$b} <=> $co{$a} : lc($a) cmp lc($b) } keys %cfs;
			print "\n" unless ($ps);
			print join(' ',' ' x 10,map { sprintf('%8s',$_) } @cfs)."\n";
			print join(' ',' ' x 10,map { '-' x 8 } @cfs)."\n";
			foreach my $cs (sort { $so{$b} <=> $so{$a} } @css) {
				print join(' ',
					sprintf('%-10s',$sn{$cs} ? "$sn{$cs}:" : "$cs:"),
					map { defined($info{counters}->{$cs}->{$_}) ? sprintf('%8u',$info{counters}->{$cs}->{$_}) : sprintf('%8s','-') } @cfs)."\n";
			}
			$ps ++;
			$pa ++;
		}
	}
	return $pa;
}

sub print_data {
	my $pa = 0;
	$pa += print_info();	
	$pa += print_counters();
	print "\n" if ($pa);
}

sub run_dig {
	my $res = query_daemon_status('OPT','ping');
	return $res unless ($res);
	my @cmd = (
		'dig' => (map { "\@$_" } @{$res->{a}}),
		'-p', => $res->{p},
		'+edns=0',
	);
	push @cmd, @ARGV ? @ARGV : ('-t','ANY',"dig$ctl->{domain}");
	@ARGV = ();
	system(@cmd);
}

sub run_drill {
	my $res = query_daemon_status('OPT','ping');
	return $res unless ($res);
	my @cmd = (
		'drill' => (map { "\@$_" } @{$res->{a}}),
		'-p', => $res->{p},
		'-D','-a',
	);
	push @cmd, @ARGV ? @ARGV : ('-t','ANY',"drill$ctl->{domain}");
	@ARGV = ();
	system(@cmd);
}

sub run_query {
	my ($qtype,$dname) = @_;
	print sprintf("%-11s %-37s %6s %s\n",'query:',$dname,'',$qtype);
	my $res = query_daemon_status($qtype,'query',$dname);
	return $res unless ($res);
	my $c = 0;
	foreach my $sec (('question','answer','authority','additional')) {
		my @inf = ();
		eval { @inf = $res->{r}->$sec; };
		foreach my $inf (@inf) {
			my $dsp = $inf->string;
			$dsp =~ s/[ \t]+;.*$//mg;
			$dsp =~ s/\s*[\r\n]+;/;/gs;
			$dsp =~ s/[\r\n]+\s*/ /gs;
			$dsp =~ s/[;\s\r\n]+$//s;
			$dsp =~ s/\s+;/;/gs;
			my $pre = '';
			if ($dsp =~ /^;\s*(.*?)\s*$/) {
				$dsp = $1;
			} elsif ($dsp =~ /^(\S+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S.*?)\s*$/) {
				$dsp = $5;
				$pre = sprintf('%-37s %6u %s',$1,$2,$4);
			} elsif ($dsp =~ /^(\S+)\s+(\S+)\s+(\S+)\s*$/) {
				$pre = sprintf('%-37s %6s %s',$1,'',$3);
				$dsp = '';
			}
			$dsp =~ s/\s+/ /gs;
			if ($dsp =~ /\S/) {
				print sprintf("%-11s %-51s %s\n","$sec:",$pre,$dsp);
			} elsif ($pre =~ /\S/) {
				print sprintf("%-11s %s\n","$sec:",$pre);
			}
			$c ++;
		}
	}
	return $c;
}

my $did = 0;
while (@ARGV) {
	my $p = shift @ARGV;
	if ($p =~ /^-*dig?$/i) {
		run_dig();
		$did ++;
	} elsif ($p =~ /^-*drill?$/i) {
		run_drill();
		$did ++;
	} elsif ($p =~ /^-*d(e?bu?g)?$/i) {
		$debug = 1;
	} elsif (@ARGV && $p =~ /^[a-zA-Z0-9]+$/) {
		run_query(uc($p),shift @ARGV);
		$did ++;
	} elsif ($p =~ /^[a-zA-Z0-9].*\./) {
		run_query('ANY',$p);
		$did ++;
	}
}

unless ($did) {
	get_meminfo();
	get_daemoninfo();
	print_data();
}

exit($cerr);
