#!/usr/bin/perl
#<: 'wtdnssign.pl' v0.003 @ 2012-10-16 16:44:44 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>#

require 'config.pl';
my %ctl = ();
$logtxt	= sub {
	my ($tag,@txt) = @_;
	print sprintf("%-10s %s\n","$tag:",join(' ',@txt));
};

require 'update.pl';
require 'control.pl';
require 'dnssec.pl';

unless (can_control()) {
	print STDERR "$ctlerr\n";
	exit(1);
}

sub sign_things {
	my ($key,@things) = @_;
	my %zones = ();
	foreach my $thing (@things) {
		my $qres = query_daemon_rr('PTR',"*.$thing");
		foreach my $qans ($qres ? $qres->answer : ()) {
			next unless ($qans->type eq 'PTR');
			if ($qans->ptrdname =~ /^[^.]*\.$thing$/) {
				$zones{$thing} = undef;
				last;
			}
		}
		next if ((exists $zones{$thing}) && !$zones{$thing});
		my $zone = $thing;
		$zone =~ s/^[^.]*//;
		$zone =~ s/^\.*//;
		next if ((exists $zones{$zone}) && !$zones{$zone});
		$zones{$zone} = [] unless ($zones{$zone});
		push @{$zones{$zone}}, $thing;
	}
	my $tc = 0;
	foreach my $zone (keys %zones) {
		my $rrs;
		if ($zones{$zone}) {
			$logtxt->('Names',$zone);
			$rrs = dnssec_sign_names($zone,$key,$zones{$zone});
		} else {
			$logtxt->('Zone',$zone);
			$rrs = dnssec_sign_zone($zone,$key);
		}
		if ($rrs) {
			my ($c,$e) += dnssec_save_rrs($zone,$key,$rrs);
			print STDERR "$e\n" if ($e);
			$tc += $c;
		}
		$rrs = dnssec_create_nsecs($zone);
		if ($rrs) {
			my ($c,$e) += dnssec_save_rrs($zone,$key,$rrs);
			print STDERR "$e\n" if ($e);
			$tc += $c;
		}
	}
	return $tc;
}

my @names = ();
my @keyl = ();
while (@ARGV) {
	my $p = shift @ARGV;
	if ($p =~ /-+K(?:eys?)?$/i) {
		push @keyl, shift @ARGV;
	} elsif (($p =~ /[\\\/*?]/) || ($p =~ /\.\+\d+\+\d+\.private$/i)) {
		push @keyl, $p;
	} elsif ($p =~ /^[^-].*\./) {
		push @names, lc($p);
	}
}
my $count = sign_things(\@keyl,@names);
print "# $count\n";
