//<: 'dnsdata.h' v0.009 @ 2012-10-08 00:50:18 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS DNS types and consts */

#include <time.h>

// Resource record identifiers

#define	REQ_A		1
#define	REQ_NS		2
#define	REQ_CNAME	5
#define	REQ_SOA		6
#define	REQ_PTR		12
#define	REQ_HINFO	13
#define	REQ_MX		15
#define	REQ_TXT		16
#define	REQ_AAAA	28
#define	REQ_SRV		33
#define	REQ_A6		38
#define REQ_OPT         41
#define REQ_DS          43
#define REQ_RRSIG       46
#define REQ_NSEC        47
#define REQ_DNSKEY      48
#define	REQ_SPF		99
#define	REQ_ANY		255
#define RECOGNIZED_RRS  18
#define SUPPORTED_RRS   RECOGNIZED_RRS-2
#define AUTOADD_RRS	2

#define	SREQ_A		"A"
#define	SREQ_NS		"NS"
#define	SREQ_CNAME	"CNAME"
#define	SREQ_SOA	"SOA"
#define	SREQ_PTR	"PTR"
#define	SREQ_HINFO	"HINFO"
#define	SREQ_MX		"MX"
#define	SREQ_TXT	"TXT"
#define	SREQ_AAAA	"AAAA"
#define	SREQ_SRV	"SRV"
#define	SREQ_A6		"A6"
#define SREQ_OPT        "OPT"
#define SREQ_DS         "DS"
#define SREQ_RRSIG      "RRSIG"
#define SREQ_NSEC       "NSEC"
#define SREQ_DNSKEY     "DNSKEY"
#define	SREQ_SPF	"SPF"
#define	SREQ_ANY	"ANY"

#define CLASS_INET      1
#define CLASS_ANY       255

#define OP_QUERY        0
#define OP_IQUERY       1
#define OP_STATUS       2

#define RC_NOERROR      0
#define RC_FORMERR      1
#define RC_SERVFAIL     2
#define RC_NXDOMAIN     3
#define RC_NOTIMP       4
#define RC_REFUSE       5

#define OPCODE          0x7800
#define RCODE           0x000F

// DNSSEC Authentic Data
#define FB_AD           0x0010
// DNSSEC Checking Disabled
#define FB_CD           0x0020
// Recursion Available
#define FB_RA           0x0080
// Recursion Desired
#define FB_RD           0x0100
// Truncated
#define FB_TC           0x0200
// Authoritative Answer
#define FB_AA           0x0400
// Query Response
#define FB_QR           0x8000
// DNSSEC OK
#define OPTFB_DO        0x8000
#define OPTFB_BAD       0x0000

#define rcode_ok(f)	(((f & RCODE) == RC_NOERROR) || ((f & RCODE) == RC_NXDOMAIN))
#define rcode_err(f)	(((f & RCODE) == RC_FORMERR) || ((f & RCODE) == RC_SERVFAIL) || ((f & RCODE) == RC_REFUSE))

#define TTL_POLICY(q,n)	(((q) == REQ_SOA || (q) == REQ_NS) ? srvttl_time : ddns_domain_cmp((char *)n) ? defttl_time : dynttl_time)

typedef struct {
  int qtype;
  char *name;
} qtypename;

/* Supported query types */
extern const int qtypes_supported[SUPPORTED_RRS];
/* Auto-add query types */
extern const int qtypes_autoadd[AUTOADD_RRS];
/* Query type name map */
extern const qtypename qtypes_names[RECOGNIZED_RRS+1];

// TTLs
extern u4 srvttl_time;
extern u4 defttl_time;
extern u4 dynttl_time;

// Do we support this query type?
int qtype_supported(int qtype);
// Get index for query type
int qtype_index(int qtype);
// Get mnemonic name for query type
char* qtype_mname(int qtype);
// Get file name for query type
char* qtype_fname(int qtype);
// Get query type for file name
int qname_type(char *name);


/* DNS trecord outputs: */
// Output DNS name
int put_name(unsigned char *buf, unsigned char **po, unsigned char *name, unsigned char *end);
// Output text
int put_text(u1 *buf, u1 **po, char *text, u1 *end);
// Output binary data from hexadecimal string
int put_hex(u1 *buf, u1 **po, char *data, u1 *end);
// Output A (IPv4) record
int put_A(u1 *buf, u1 **po, char *addr, u1 *end);
// Output AAAA (IPv6) record
int put_AAAA(u1 *buf, u1 **po, u2 type, char **addrsf, char *addr, u1*end);
// Output A6 (IPv6 with prefix) record
int put_A6(u1 *buf, u1 **po, u2 type, char **qres, char *addr, u1*end);
// Output SOA (start of Authority) record
int put_SOA(u1 *buf, u1 **po, char **qres, char *addrs, struct tm *tm, u1*end);
// Output MX (mail exchange) record, with counter based priority
int put_MX(u1 *buf, u1 **po, char *addr, u2 count, u1*end);
// Output SRV (service) record, with counter based priority
int put_SRV(u1 *buf, u1 **po, char *addr, u2 count, u1*end);
// Output HINFO (host info) record
int put_HINFO(u1 *buf, u1 **po, char *info, u1*end);
// Output OPT (EDNS pseudo) record
int put_OPT(u1 *buf, u1 **po, u1 optrcode, u2 optflags, u1 *end);
