//<: 'dnssec.c' v0.011 @ 2012-10-08 21:22:10 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
 
#include "b64/cdecode.h"

#include "knobs.h"
#include "global.h"
#include "dnssec.h"
#include "dnsdata.h"
#include "dir.h"
#include "strutil.h"

/* What the DNS DNSSEC stuff */

// Output binary data from base64 encoded string
int put_base64(u1 *buf, u1 **po, char *data, u1 *end)
{
  unsigned long dl = strlen(data);
  u1 *bin = malloc(dl);
  base64_decodestate b64ds;
  base64_init_decodestate(&b64ds);
  int bl = base64_decode_block(data, dl, (char *)bin, &b64ds);
  if (bl < 1) {
    free(bin);
    return -2;
  }
  for (int i = 0; i < bl; i++)
  {
    if (*po >= end)
    {
      free(bin);
      return -1;
    }
    *(*po)++ = bin[i];
  }
  free(bin);
  return 0;
}

// Output RRSIG (resource record signature) record
int put_RRSIG(u1 *buf, u1 **po, char **name, u2 type, char *rr, u1 *end)
{
  if (*po + 18 >= end)
    return -1;
  char *s, *e;
  u4 ull;
  struct tm tm;
  int ni, rri = 0;
  // Get algorithm (integer)
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 alg = ull;
  // Get labels (integer)
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 labc = ull;
  // Get original TTL (integer)
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) >= 0) && !*e))
    return -2;
  u4 ottl = ull;
  // Get expiration time (YYYYMMDDhhmmss)
  memset(&tm, 0, sizeof(tm));
  ni = 0;
  if (!((s = next_token(rr, &rri, 0)) && *s &&
        (tm.tm_year = next_number(s, &ni, 4)) && (ni > 0) &&
        (tm.tm_mon  = next_number(s, &ni, 2)) && (ni > 0) &&
        (tm.tm_mday = next_number(s, &ni, 2)) && (ni > 0) &&
        ((tm.tm_hour = next_number(s, &ni, 2)) < 26) && (ni > 0) &&
        ((tm.tm_min  = next_number(s, &ni, 2)) < 61) && (ni > 0) &&
        ((tm.tm_sec  = next_number(s, &ni, 2)) < 61) && (ni > 0)))
    return -2;
  u4 expt = timegm(&tm);
  // Get inception time (YYYYMMDDhhmmss)
  memset(&tm, 0, sizeof(tm));
  ni = 0;
  if (!((s = next_token(rr, &rri, 0)) && *s &&
        (tm.tm_year = next_number(s, &ni, 4)) && (ni > 0) &&
        (tm.tm_mon  = next_number(s, &ni, 2)) && (ni > 0) &&
        (tm.tm_mday = next_number(s, &ni, 2)) && (ni > 0) &&
        ((tm.tm_hour = next_number(s, &ni, 2)) < 26) && (ni > 0) &&
        ((tm.tm_min  = next_number(s, &ni, 2)) < 61) && (ni > 0) &&
        ((tm.tm_sec  = next_number(s, &ni, 2)) < 61) && (ni > 0)))
  {
    return -2;
  }
  u4 inct = timegm(&tm);
  // Get key tag (integer)
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x10000) && !*e))
    return -2;
  u2 tag = ull;
  // Get key owner name (string)
  if (!((s = next_token(rr, &rri, 0)) && *s))
    return -2;
  *name = s;
  // Get signature (base64 binary)
  if (!((s = next_token(rr, &rri, 0)) && *s))
    return -2;
  *(*po)++ = type >> 8;
  *(*po)++ = type & 0xFF;
  *(*po)++ = alg;
  *(*po)++ = labc;
  *(*po)++ = (ottl >> 24) & 0xFF;
  *(*po)++ = (ottl >> 16) & 0xFF;
  *(*po)++ = (ottl >> 8) & 0xFF;
  *(*po)++ = (ottl >> 0) & 0xFF;
  *(*po)++ = (expt >> 24) & 0xFF;
  *(*po)++ = (expt >> 16) & 0xFF;
  *(*po)++ = (expt >> 8) & 0xFF;
  *(*po)++ = (expt >> 0) & 0xFF;
  *(*po)++ = (inct >> 24) & 0xFF;
  *(*po)++ = (inct >> 16) & 0xFF;
  *(*po)++ = (inct >> 8) & 0xFF;
  *(*po)++ = (inct >> 0) & 0xFF;
  *(*po)++ = tag >> 8;
  *(*po)++ = tag & 0xFF;
  int np = put_name(buf, po, (u1 *)(*name), end);
  if (np < 0)
    return np;
  ni = put_base64(buf, po, s, end);
  if (ni < 0)
    return ni;
  return np;
}

// Output DNSKEY (DNSSEC public key) record
int put_DNSKEY(u1 *buf, u1 **po, char *rr, u1 *end)
{
  if (*po + 4 >= end)
    return -1;
  char *s, *e;
  u4 ull;
  int rri = 0;
  // Get flags
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x10000) && !*e))
    return -2;
  u2 flags = ull;
  // Get protocol
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 prot = ull;
  // Get algorithm
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 alg = ull;
  // Get key
  if (!((s = next_token(rr, &rri, 0)) && *s))
    return -2;
  *(*po)++ = flags >> 8;
  *(*po)++ = flags & 0xFF;
  *(*po)++ = prot;
  *(*po)++ = alg;
  return put_base64(buf, po, s, end);
}

// Output DS (DNSSEC delegation signer) record
int put_DS(u1 *buf, u1 **po, char *rr, u1 *end)
{
  if (*po + 4 >= end)
    return -1;
  char *s, *e;
  u4 ull;
  int rri = 0;
  // Get tag
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x10000) && !*e))
    return -2;
  u2 tag = ull;
  // Get algorithm
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 alg = ull;
  // Get digest type
  if (!((s = next_token(rr, &rri, 0)) && *s && ((ull = strtoul(s, &e, 10)) < 0x100) && !*e))
    return -2;
  u1 digt = ull;
  // Get digest
  if (!((s = next_token(rr, &rri, 0)) && *s))
    return -2;
  *(*po)++ = tag >> 8;
  *(*po)++ = tag & 0xFF;
  *(*po)++ = alg;
  *(*po)++ = digt;
  return put_hex(buf, po, s, end);
}

// Output NSEC (DNSSEC next secure) record
int put_NSEC(u1 *buf, u1 **po, char *rr, u1 *end)
{
  int ni = 0;
  // Get next name
  char *next = next_token(rr, &ni, 0);
  if (!next)
    return -2;
  // Get type list
  char *type = next_token(rr, &ni, 0);
  if (!type)
    return -2;
  // Get type bitmap
  char *bitmap = next_token(rr, &ni, 0);
  if (*type && !bitmap) // Bitmap can be missing if type list is empty
    return -2;
  int np = put_name(buf, po, (u1 *)next, end);
  if (np < 0)
    return np;
  if (bitmap && *bitmap) // For now we can only work with ready made bitmap
    ni = put_hex(buf, po, bitmap, end);
  else
    for (ni = 0; ni < 8; ni++)
      *(*po)++ = 0;
  if (ni < 0)
    return ni;
  return np;
}

// Try to find NSEC records:
int find_nsecs(char **psec0, char **psec1, unsigned char *query, int flags)
{
  (*psec0) = NULL;
  (*psec1) = NULL;
  char *mine[2];
  char *prev[2];
  prev[0] = NULL;
  prev[1] = NULL;
  char *q = (char *)query;
  while (*q && (*q != '.'))
    q++;
  int gc = 0;
  if (q > ((char *)query))
  {
    mine[0] = strndup((char *)query, q - ((char *)query));
    int hc = 1;
    if (!(flags & FNS_WILD) || (*query == '*'))
      mine[1] = NULL;
    else
    {
      mine[1] = "*";
      hc++;
    }
    while (*q && (*q == '.'))
      q++;
    if (search_for_hosts_in_domain(NULL, NULL, mine, NULL, prev, hc, q, SDH_INC_PSEUDO))
    {
      int pl;
      int ql = strlen(q);
      if (prev[0])
      {
        pl = strlen(prev[0]);
        if (pl & ql)
        {
          (*psec0) = malloc(pl + ql + 2);
          strncpy((*psec0), prev[0], pl);
          (*psec0)[pl] = '.';
          strncpy((*psec0)+pl+1, q, ql+1);
        }
        else if (pl)
          (*psec0) = strdup(prev[0]);
        else if (ql)
          (*psec0) = strdup(q);
        else
          (*psec0) = strdup("");
        gc++;
        free(prev[0]);
      }
      if (prev[1])
      {
        pl = strlen(prev[1]);
        if (pl & ql)
        {
          (*psec1) = malloc(pl + ql + 2);
          strncpy((*psec1), prev[1], pl);
          (*psec1)[pl] = '.';
          strncpy((*psec1)+pl+1, q, ql+1);
        }
        else if (pl)
          (*psec1) = strdup(prev[1]);
        else if (ql)
          (*psec1) = strdup(q);
        else
          (*psec1) = strdup("");
        gc++;
        free(prev[1]);
      }
    }
  }
  free(mine[0]);
  return gc;
}

