//<: 'global.c' v0.007 @ 2012-10-08 18:55:38 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS global data */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#include "knobs.h"
#include "global.h"
#include "strutil.h"
#include "log.h"
#include "dir.h"
#include "dnsdata.h"

char *program = "";
char *versionstring = "";
char *usagestring = "";

int             hash_type       = HASH_NEW;
char           *data_dir        = DATA_DIR;
int             dsec_on         = 0;
int             dsec_auth       = 0;
int             dsec_ddns_auth  = 0;
int             ddns_timeouts   = 0;
int             ddns_usergroup  = 0;
char           *ddns_domain     = NULL;
int             ddns_domain_len = 0;
unsigned int    ddns_purge      = 0;
unsigned int    ddns_fallback   = 0;
uid_t           ddns_user       = 0;
gid_t           ddns_group      = 0;

// Check standard command line boolean options
int global_arg_option(char *opt)
{
  int ht = hash_type;
  if (!strncmp(opt, "-oh", 4))
  {
    hash_type = HASH_OLD;
    ht = hash_type;
  }
  else if (!strncmp(opt, "-uh", 4))
  {
    if (hash_type == HASH_NO)
      hash_type = (ht == HASH_NO) ? HASH_NEW : ht;
  }
  else if (!strncmp(opt, "-nh", 4))
    hash_type = HASH_NO;
  else if (!strncmp(opt, "-ds", 4))
    dsec_on ++;
  else if (!strncmp(opt, "-ad", 4))
    dsec_auth ++;
  else if (!strncmp(opt, "-ddad", 6))
    dsec_ddns_auth ++;
  else if (!strncmp(opt, "-ddto", 6))
    ddns_timeouts ++;
  else if (!strncmp(opt, "-ddug", 6))
    ddns_usergroup ++;
  else if (!strncmp(opt, "-ls", 4)) 
  {
    if (loglevel < LOG_WARNING)
      loglevel = LOG_WARNING;
    syslogopen(program);
  }
  else if (!strncmp(opt, "-v", 3))
    loglevel ++;
  else if (!strncmp (opt, "-V", 3))
  {
    fprintf (stdout, "%s (%s)\n", versionstring, MAKEDATE);
    exit(0);
  }
  else if (!strncmp (opt, "-h", 3))
  {
    fprintf (stdout, "%s", usagestring);
    exit(0);
  }
  else
    return 0;
  return 1;
}

// Check standard command line settings
int global_arg_setting(char *opt, char *val)
{
  if (!strncmp(opt, "-lf", 4))
  {
    if (loglevel < LOG_NOTICE)
      loglevel = LOG_NOTICE;
    logopen(val);
  }
  else if (!strncmp(opt, "-ld", 4))
  {
    if (loglevel < LOG_DEBUG)
      loglevel = LOG_DEBUG;
    dbglogopen(val);
  }
  else if (!strncmp(opt, "-ll", 4))
    loglevel = atoi(val);
  else if (!strncmp(opt, "-cf", 4))
  {
    if (read_config(val))
      logmsg(LOG_INFO,"Read config: %s",val);
    else
      logerr("Config file '%s'",val);
  }
  else if (!strncmp(opt, "-dd", 4))
  {
    data_dir = val;
  }
  else
    return 0;
  return 1;
}

// Parse a bool setting
static int strbool(char *s, int d)
{
  if (!(s && *s))
    return d;
  if (!(strncasecmp(s, "yes", 4) && strncasecmp(s, "on", 3) && strncasecmp(s, "true", 5)))
    return 1;
  if (!(strncasecmp(s, "no", 3) && strncasecmp(s, "off", 4) && strncasecmp(s, "false", 6)))
    return 0;
  char *e;
  long long int i = strtoll(s, &e, 10);
  return *e ? d : i ? 1 : 0;
}

#define INI_GLOBAL      0
#define INI_FALLBACKS   'F'
#define INI_OTHER       '?'

struct ddnsfallback {
    struct ddnsfallback *next;
    int type;
    char *value;
};
static struct ddnsfallback ddnsfallbacks = { NULL, 0, NULL };

char *ddns_fallback_record(int type)
{
  struct ddnsfallback *dfb = ddnsfallbacks.next;
  while (dfb)
  {
    if (dfb->type == type)
      return dfb->value;
    dfb = dfb->next;
  }
  return NULL;
}

int ddns_domain_cmp(char *name)
{
  if (!ddns_domain_len)
    return 1;
  int l = strlen(name);
  if (l < ddns_domain_len)
    return -1;
  return strncmp(name + (l - ddns_domain_len), ddns_domain, ddns_domain_len);
}

// Read and parse configuration file
int read_config(char *file)
{
  char buf[1024];
  char *l;
  long int i;
  FILE *ini = fopen(file, "r");
  if (!ini)
    return 0;
  int oh = (hash_type == HASH_OLD);
  int uh = (hash_type != HASH_NO);
  char *dds = data_dir;
  char inisec = INI_GLOBAL;
  while ((l = fgets(buf, sizeof(buf), ini)))
  {
    if (*l && l[strlen(l) - 1] != '\n')
    {
        while ((l = fgets(buf, sizeof(buf), ini)) && *l && (l[strlen(l) - 1] != '\n'));
        continue;
    }
    string_chomp((u1 *)l);
    if ((*l == ';') || (*l == '#'))
      continue;
    char *e = l + (strlen(l) - 1);
    if ((*l == '[') && (*e == ']'))
    {
      l++;
      *e-- = 0;
      string_chomp((u1 *)l);
      lower_case_name((u1 *)l);
      if (!strncmp(l, "fallbacks", 10))
        inisec = INI_FALLBACKS;
      else
        inisec = INI_OTHER;
      continue;;
    }
    if (inisec == INI_OTHER)
      continue;
    char *o = strsep(&l, "=");
    if (o && l)
    {
      while (isspace(*o))
        o++;
      e = l - 2;
      while (e > o && isspace(*e))
        e--;
      *(++e) = 0;
      while (isspace(*l))
        l++;
      e = l + (strlen(l) - 1);
      while (e > l && isspace(*e))
        e--;
      *(++e) = 0;
      switch(inisec)
      {
        case INI_GLOBAL:
          lower_case_name((u1 *)o);
          if (!strncmp(o, "hashing", 8))
            uh = strbool(l, uh);
          else if (!strncmp(o, "oldhash", 8))
            oh = strbool(l, oh);
          else if (!strncmp(o, "datadir", 8))
          {
            if (dds != data_dir)
              free(data_dir);
            data_dir = strndup(l, 255);
          }
          else if (!strncmp(o, "domain", 7))
          {
            if (ddns_domain)
              free(ddns_domain);
            string_clean_resource((u1 *)l);
            lower_case_name((u1 *)l);
            if (!*l || (*l == '.'))
            {
              ddns_domain = strndup(l, 255);
              ddns_domain_len = strlen(ddns_domain);
            }
            else
            {
              i = strlen(l);
              if (i > 254)
                i = 254;
              ddns_domain = malloc(i + 2);
              *ddns_domain = '.';
              strncpy(ddns_domain+1, l, i);
              ddns_domain[i+1] = 0;
              ddns_domain_len = i+1;
            }
          }
          else if (!strncmp(o, "purge", 6))
            ddns_purge = parse_duration(l);
          else if (!strncmp(o, "fallback", 9))
            ddns_fallback = parse_duration(l);
          else if (!strncmp(o, "user", 9))
            ddns_user = get_user_id(l, ddns_user);
          else if (!strncmp(o, "group", 9))
            ddns_group = get_group_id(l, ddns_group);
          break;
        case INI_FALLBACKS:
          string_clean_resource((u1 *)o);
          upper_case_name((u1 *)o);
          i = qname_type(o);
          if (i > 0)
          {
            struct ddnsfallback *dfb = ddnsfallbacks.next;
            while (dfb)
            {
              if (dfb->type == i)
              {
                free(dfb->value);
                break;
              }
              dfb = dfb->next;
            }
            if (!dfb)
            {
              dfb = (struct ddnsfallback *)malloc(sizeof(struct ddnsfallback));
              dfb->type = i;
              dfb->next = ddnsfallbacks.next;
              ddnsfallbacks.next = dfb;
            }
            dfb->value = strndup(l, 1024);
          }
          break;
      }
    }
  }
  fclose(ini);
  if (uh)
    hash_type = oh ? HASH_OLD : HASH_NEW;
  else
    hash_type = HASH_NO;
  return 1;
}

#define nnfree(p) do { if (p) { free(p); p=NULL; } } while (0)

void free_globals()
{
  nnfree(ddns_domain);
  struct ddnsfallback *dfb = ddnsfallbacks.next;
  while (dfb)
  {
    struct ddnsfallback *dfbt = dfb;
    nnfree(dfb->value);
    dfb = dfb->next;
    free(dfbt);
  }
  ddnsfallbacks.next = NULL;
}

uid_t get_user_id(char *name, uid_t dusr)
{
  if (!*name)
    return 0;
  char *e = name;
  int i = strtol(name, &e, 10);
  if (e && !*e && (i >= 0))
    return i;
  struct passwd *pwd = getpwnam(name);
  if (pwd)
    return pwd->pw_uid;
  return dusr;
}

gid_t get_group_id(char *name, gid_t dgrp)
{
  if (!*name)
    return 0;
  char *e = name;
  int i = strtol(name, &e, 10);
  if (e && !*e && (i >= 0))
    return i;
  struct group *grp = getgrnam(name);
  if (grp)
    return grp->gr_gid;
  return dgrp;
}
