//<: 'strutil.c' v0.025 @ 2012-10-08 20:51:35 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS string stuff */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <sys/utsname.h>

#include "knobs.h"
#include "global.h"
#include "strutil.h"
#include "log.h"

void string_clean_text(unsigned char *p)
{
  unsigned char *q;
  int n;
  assert(p);
  for (n = 0, q = p; *q && n < MAX_RECORD_LEN; q++, n++)
    if (!bad_text_char(*q))
      *p++ = *q;
  *p = '\0';
}

void string_clean_resource(unsigned char *p)
{
  unsigned char *q;
  int n;
  assert(p);
  for (n = 0, q = p; *q && n < MAX_RECORD_LEN; q++, n++)
    if (!bad_resource_char(*q))
      *p++ = *q;
  *p = '\0';
}

#define lower_case_name_char(c) (c | 0x20)
void lower_case_name(unsigned char *n)
{
  assert(n);
  for (; *n; n++)
    *n = lower_case_name_char(*n);
}

#define upper_case_name_char(c) (c & ~0x20)
void upper_case_name(unsigned char *n)
{
  assert(n);
  for (; *n; n++)
    *n = upper_case_name_char(*n);
}

void string_chomp(unsigned char *p)
{
  unsigned char *q, *t;
  assert (p);
  /* skip over leading spaces */
  for (q = p; *q; q++)
    if (!isspace(*q))
      break;
    /* skip over non-space */
  for (t = p; *q;)
    if (!isspace (*p++ = *q++))
      t = p;
  /* truncate trailing space */
  *t = '\0';
}

char **string_split(char *in, char c, int maxrec, int maxbuf, int maxline)
{
  if (!in)
    return NULL;
  while (*in == c)
    in ++;
  if (!(*in))
    return NULL;
  if (maxrec < 0)
    maxrec = 256;
  if (maxbuf < 0)
    maxbuf = MAX_BUF;
  if (maxline < 0)
    maxline = MAX_RECORD_LEN;
  int l, z;
  int rc = 0, ri = 0, bz = 0;
  char *po, *pi, *pc;
  pi = in;
  while (pi && *pi)
  {
    l = 0;
    pc = pi;
    while (*pi && (*pi != c))
    {
      pi++;
      l++;
    }
    if ((l > 0) && ((maxline < 1) || (l <= maxline)))
    {
      z = sizeof(char *) + l + 1;
      if ((maxbuf < 1) || (bz + z <= maxbuf))
      {
        rc ++;
        bz += z;
        if ((maxrec > 0) && (rc >= maxrec))
          break;
      }
    }
    while (*pi && *pi == c)
      pi++;
  }
  if (!(rc && bz))
    return NULL;
  bz += sizeof(char *);
  char **lst = malloc(bz);
  lst[rc] = 0;
  if (!lst)
    return NULL;
  bz = 0;
  pi = in;
  po = (char *)&(lst[rc+1]);
  ri = 0;
  while (pi && *pi)
  {
    l = 0;
    pc = pi;
    while (*pi && (*pi != c))
    {
      pi++;
      l++;
    }
    if ((l > 0) && ((maxline < 1) || (l <= maxline)))
    {
      z = sizeof(char *) + l + 1;
      if ((maxbuf < 1) || (bz + z <= maxbuf))
      {
        lst[ri++] = po;
        bz += z;
        strncpy(po, pc, l);
        po += l;
        *(po++) = 0;
        if (ri >= rc)
          break;
      }
    }
    while (*pi && *pi == c)
      pi++;
  }
  return lst;
}

///* this is quite possibly my favorite function in the whole world */
//char **string_split (const char *s, char c, size_t max, int multi)
//{
//  char *p, **a;
//  int n;
//  for (n = 0, p = (char *) s; *p && n < max; n++)
//  {
//    while (*p && *p != c)
//      p++;
//    if (multi)
//    {
//      while (*p && *p == c)
//        p++;
//    }
//    else if (*p)
//      p++;
//  }
//  if (!(a = malloc ((n + 1) * sizeof (char *) + strlen (s) + 1)))
//    return NULL;
//  p = (char *) (&(a[n + 1]));
//  for (n = 0; *s && n < max; n++)
//  {
//    a[n] = p;
//    while (*s && *s != c)
//      *p++ = *s++;
//    if (multi)
//    {
//      *p = '\0';
//      while (*s && *s == c)
//      {
//        *p++ = '\0';
//	s++;
//      }
//    }
//    else if (*s)
//    {
//      *p++ = '\0';
//      if (*s == c)
//        s++;
//    }
//    else
//    {
//      *p = '\0';
//    }
//  }
//  a[n] = 0;
//  return a;
//}

int string_present(char *s, char **a)
{
  int i;
  for (i = 0; a[i]; i++)
    if (!strcmp (a[i], s))
      return 1;
  return 0;
}

typedef struct {
  char *str;
  int mult;
} timeunit;

#define MAX_TIME_UNITS 17
#define FMT_TIME_UNITS  6
const timeunit timeunits[MAX_TIME_UNITS] = {
  {"s",1},
  {"m",60},
  {"h",3600},           //  60      *    60
  {"d",86400},          //  24      *  3600
  {"m",2629800},        //  30.4375 * 86400
  {"y",31557600},       // 365.25   * 86400
  {"w",604800},         //   7      * 86400
  {"sec",1},
  {"min",60},
  {"mon",2629800},
  {"second",1},
  {"minute",60},
  {"hour",3600},
  {"day",86400},
  {"week",604800},
  {"month",2629800},
  {"year",31557600}
};

unsigned int parse_duration(char *s)
{
  int r = 0;
  int l,i;
  char *p;
  int v = 0;
  char u[8] = "";
  while (s && *s && (sscanf(s, "%d%7s%n", &v, u, &l) > 0))
  {
    if (u[0])
    {
      p = u;
      for (; *p; p++)
        *p = lower_case_name_char(*p);
      if (*--p == 's')
        *p = 0;
      for (i = 0; i < MAX_TIME_UNITS; i++)
      {
        if (!strncmp(timeunits[i].str, u, 7))
        {
          v *= timeunits[i].mult;
          break;
        }
      }  
    }
    r += v;
    if (l <= 0)
      break;
    while (*s && l)
    {
      s++;
      l--;
    }
    v = 0;
    u[0] = 0;
  }
  return abs(r);
}

int format_duration(char *buf, int len, int secs)
{
  int c = 0;
  for (int i = FMT_TIME_UNITS - 1; (secs > 0) && (i >= 0); i--)
  {
    int v = secs / timeunits[i].mult;
    secs %= timeunits[i].mult;
    if (v > 0)
    {
      int r = snprintf(buf+c, len-c, "%d%s ", v, timeunits[i].str);
      c += r;
      if (c >= len)
        return c;
    }
  }
  string_chomp((u1 *)buf);
  return c;
}

int format_time_since(char *buf, int len, time_t then)
{
  time_t now = time(NULL);
  int c = strftime(buf, len, "%F %T UTC - ", gmtime(&then));
  if (c < 1)
    return len + 1;
  if ((c >= len) || (now <= then))
    return c;
  int r = strftime(buf+c, len-c, "%F %T UTC   ", gmtime(&now));
  if (r < 1)
    return len + 1;
  c += r;
  if (c >= len)
    return c;
  r = format_duration(buf+c, len-c, now - then);
  if (r > 0)
    buf[c-2] = '=';
  c += r;
  if (c >= len)
    return c;
  string_chomp((u1 *)buf);
  return c;
}

static const char alnum[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-";

int random_name(char *buf, int len)
{
  if (len > 1)
  {
    int ll = 1;
    buf[0] = alnum[random() % (sizeof(alnum) - 3)];
    for (int i=1; i < len-1; i++)
    {
      if ((buf[i-1] == '-') || (buf[i-1] == '.'))
      {
        buf[i] = alnum[random() % (sizeof(alnum) - 3)];
        ll++;
      }
      else if (ll >= 32)
      {
        buf[i] = '.';
        ll = 0;
      }
      else
      {
        buf[i] = alnum[random() % (sizeof(alnum) - 1)];
        if (buf[i] == '.')
          ll = 0;
        else
          ll++;
      }
    }
    buf[len-1] = alnum[random() % (sizeof(alnum) - 3)];
  }
  buf[len] = 0;
  return len;
}

int sub_of_domain(char *name, char *domain)
{
  int dl = strlen(domain);
  int nl = strlen(name);
  int dp = nl - dl;
  if (nl < dl)
    return -1;
  if (strncmp(name+dp, domain, dl))
    return -1;
  name[dp] = 0;
  return dp;
}

char *next_token(char *tokens, int *index, char delim)
{
  if (!tokens[*index])
    return NULL;
  char *tok = tokens + *index;
  char q = 0;
  int p = 0;
  while (tokens[*index])
  {
    if (tokens[*index] == q)
    {
      if ((q != '"') && (q != '\''))
        tok[p++] = q;
      q = 0;
    }
    else if (q)
    {
      tok[p++] = tokens[*index];
    }
    else if (tokens[*index] == delim)
    {
      (*index) ++;
      break;
    }
    else if (tokens[*index] == '"' || tokens[*index] == '\'')
    {
      q = tokens[*index];
    }
    else if (tokens[*index] == '[')
    {
      q = ']';
      tok[p++] = tokens[*index];
    }
    else if (tokens[*index] == '{')
    {
      q = '}';
      tok[p++] = tokens[*index];
    }
    else if (!delim && (tokens[*index] == ':' || tokens[*index] == ';'))
    {
      (*index) ++;
      break;
    }
    else if (tokens[*index] == '\\')
    {
      (*index)++;
      if (tokens[*index])
        tok[p++] = tokens[*index];
      else
        break;
    }
    else if (isspace(tokens[*index]))
    {
      if ((p > 0) && !isspace(tok[p-1]))
        tok[p++] = ' ';
    }
    else
    {
      tok[p++] = tokens[*index];
    }
    (*index)++;
  }
  tok[p] = 0;
  while ((p > 0) && isspace(tok[p-1]))
    tok[--p] = 0;
  while (*tok && isspace(*tok))
    tok++;
  return tok;
}

int escape_string(char *dst, char *src, int sz)
{
  int p = 0;
  while (*src && isspace(*src))
    src++;
  sz--;
  while (*src && p < sz)
  {
    if ((*src == ':') || (*src == ';') || (*src == '"') || (*src == '\'') || (*src == '\\'))
    {
      if (p >= sz-1)
        break;
      dst[p++] = '\\';
    }
    dst[p++] = *src;
    src++;
  }
  while ((p > 0) && isspace(dst[p]))
    p--;
  if (p < sz)
    dst[p] = 0;
  return p;
}

char *host_info()
{
  struct utsname un;
  if (uname(&un))
    return NULL;
  char *info = malloc(512);
  int i = escape_string(info, un.machine, 255);
  info[i++] =';';
  int l = escape_string(&info[i], un.sysname, 255);
  info[i+(l++)] = ' ';
  l += escape_string(&info[i+l], un.release, 255-l);
  info[i+(l++)] = ' ';
  l += escape_string(&info[i+l], un.version, 255-l);
  return info;
}

#define next_number_error() do { if (*index > 0) (*index) = 0; (*index)--; return 0; } while(0)
u4 next_number(char *str, int *index, int max)
{
  if ((*index < 0) || !str)
    next_number_error();
  while (str[*index] && ((str[*index] < '0') || (str[*index] > '9')))
    if (((str[*index] < 'A') || (str[*index] > 'Z')) && ((str[*index] < 'a') || (str[*index] > 'z')))
      (*index)++;
    else 
      next_number_error();
  int i = *index;
  int l = 0;
  while ((!max || (l < max)) && str[*index] && ((str[*index] >= '0') || (str[*index] <= '9')))
  {
    (*index)++;
    l++;
  }
  if (!l)
    next_number_error();
  char c = str[*index];
  char *e;
  str[*index] = 0;
  u4 r = strtoull(str+i, &e, 10);
  if (*e)
    e = NULL;
  str[*index] = c;
  if (!e)
    *index = -1;
  return r;
}

